////////////////////////////////////////////////////////////////////////////////////
// $Revision: 59311 $
////////////////////////////////////////////////////////////////////////////////////
// Copyright 2016 SlickEdit Inc. 
// You may modify, copy, and distribute the Slick-C Code (modified or unmodified) 
// only if all of the following conditions are met: 
//   (1) You do not include the Slick-C Code in any product or application 
//       designed to run independently of SlickEdit software programs; 
//   (2) You do not use the SlickEdit name, logos or other SlickEdit 
//       trademarks to market Your application; 
//   (3) You provide a copy of this license with the Slick-C Code; and 
//   (4) You agree to indemnify, hold harmless and defend SlickEdit from and 
//       against any loss, damage, claims or lawsuits, including attorney's fees, 
//       that arise or result from the use or distribution of Your application.
////////////////////////////////////////////////////////////////////////////////////
#pragma option(pedantic,on)
#region Imports
#include "slick.sh"
#include "tagsdb.sh"
#import "se/tags/TaggingGuard.e"
#import "cbrowser.e"
#import "context.e"
#import "files.e"
#import "stdcmds.e"
#import "stdprocs.e"
#import "tags.e"
#endregion

/**
 * Called when this module is loaded (after definit). 
 *  
 * Sets up key binding for edit-associated-symbol, if Alt+` is not already used. 
 */
defload()
{
   key_index := event2index(name2event('a-`'));
   if ( ! eventtab_index(_default_keys,_default_keys,key_index) ) {
      set_eventtab_index(_default_keys,key_index,
                         find_index('edit_associated_symbol',COMMAND_TYPE));
   }
   if (_isMac()) {
      key_index = event2index(name2event('m-`'));
      if ( ! eventtab_index(_default_keys,_default_keys,key_index) ) {
         set_eventtab_index(_default_keys,key_index,
                            find_index('edit_associated_symbol',COMMAND_TYPE));
      }
   }
}

/** 
 * Jump to the symbol corresponding to this symbol.  For example, if the current 
 * symbol under the cursor is a function, this macro would locate the corresponding 
 * function prototype and jump to that symbol.  If the file containing that symbol 
 * is already open in the editor, the position will be left intact if it is already 
 * on the function in question. 
 *  
 * @see push_tag 
 * @see edit_associated_file
 *
 * @appliesTo Edit_Window
 * @categories Search_Functions
 */
_command void edit_associated_symbol() name_info(','VSARG2_READ_ONLY|VSARG2_EDITORCTL|VSARG2_REQUIRES_MDI|VSARG2_REQUIRES_TAGGING)
{
   // make sure the current symbol context is up-to-date
   _UpdateContext(true);

   // make sure that the context doesn't get modified by a background thread.
   se.tags.TaggingGuard sentry;
   sentry.lockContext(false);

   // get list of tag files for this context
   tag_files := tags_filenamea();
   VS_TAG_BROWSE_INFO visited:[];
   num_matches := 0;
   boolean been_there_done_that:[];
   this_one_found_at := -1;

   // find the current symbol under the cursor (the function we are in)
   VS_TAG_BROWSE_INFO options[];
   context_id := tag_current_context();
   if (context_id <= 0) {
      message("No symbol under cursor");
   }

   // find other symbols with the same name in the same class
   VS_TAG_BROWSE_INFO cm;
   tag_get_context_info(context_id, cm);
   tag_push_matches();
   tag_list_symbols_in_context(cm.member_name, cm.class_name, 
                               0, 0, tag_files, "", 
                               num_matches, def_tag_max_function_help_protos, 
                               VS_TAGFILTER_ANYTHING, VS_TAGCONTEXT_ANYTHING,
                               true, p_LangCaseSensitive, visited);

   
   // bubble-sort the matches 
   n := tag_get_num_of_matches();
   tag_get_all_matches(auto matches);
   for (i:=0; i<n; i++) {
      for (j:=i; j<n; j++) {
         if (_file_case(matches[i].file_name) > _file_case(matches[j].file_name) || matches[i].line_no > matches[j].line_no) {
            VS_TAG_BROWSE_INFO tmp = matches[i];
            matches[i] = matches[j];
            matches[j] = tmp;
         }
      }
   }

   // go through the matches and remove duplicates and non-matches.
   for (i=0; i<n; i++) {
      VS_TAG_BROWSE_INFO ci = matches[i];

      // check for duplicates
      key := _file_case(ci.file_name):+"\1":+ci.return_type:+VS_TAGSEPARATOR_args:+ci.arguments;
      if (been_there_done_that._indexin(key)) continue;
      been_there_done_that:[key] = true;

      // the symbol name and language mode must match
      if (!strieq(cm.member_name, ci.member_name)) continue;

      // the class name must be similar at least
      if (cm.class_name == "" && ci.class_name != "") continue;
      if (ci.class_name == "" && cm.class_name != "") continue;
      if (cm.class_name != "" && ci.class_name != "" && (!pos(cm.class_name,ci.class_name,1,'i') || !pos(ci.class_name,cm.class_name,1,'i'))) continue;

      // we have to have a file name, and not a .jar or .dll file.
      if (ci.file_name == "") continue;
      if (_QBinaryLoadTagsSupported(ci.file_name)) continue;
      if (!file_exists(ci.file_name)) continue;

      // make sure langauge mode matches
      if (ci.language == "") ci.language = _Filename2LangId(ci.file_name);
      if (ci.language != cm.language) continue;

      // same function as the one we started in, then skip it
      if (file_eq(ci.file_name, p_buf_name) && ci.line_no >= cm.line_no && ci.line_no <= cm.end_line_no) {
         this_one_found_at = options._length();
      }

      // add this item to the list of matches
      options[options._length()] = ci;
   }

   // no matches?
   if (options._length() <= 0 || (this_one_found_at>=0 && options._length() <= 1)) {
      message("No matching symbol found for '"cm.member_name"'");
      return;
   }

   // determine which symbol to go to next
   if (this_one_found_at < 0) {
      cm = options[options._length()-1];
      this_one_found_at = options._length()+1;
   } else if (this_one_found_at+1 >= options._length()) {
      cm = options[0];
      this_one_found_at = 1;
   } else {
      cm = options[this_one_found_at+1];
      this_one_found_at++;
   }

   // open the file associated with this symbol
   message("Jumping to next symbol match out of "options._length()+1);
   dest_line_no := cm.line_no;
   parse cm.file_name with auto dest_file_name "\1" .;
   edit(maybe_quote_filename(dest_file_name));

   do {
      // update the set of symbols for this file
      _UpdateContext(true);
      // and verify if the cursor is currently already on the symbol we want
      context_id = tag_current_context();
      if (context_id <= 0) break;
      VS_TAG_BROWSE_INFO ci;
      tag_get_context_info(context_id, ci);
      if (!strieq(cm.member_name, ci.member_name)) break;
      if (cm.class_name == "" && ci.class_name != "") break;
      if (ci.class_name == "" && cm.class_name != "") break;
      if (cm.class_name != "" && ci.class_name != "" && (!pos(cm.class_name,ci.class_name,1,'i') || !pos(ci.class_name,cm.class_name,1,'i'))) break;
      if (cm.arguments != ci.arguments) break;
      if (cm.return_type != ci.return_type) break;

      // same function, we are right where we want to be
      if (p_RLine >= ci.line_no && p_RLine <= ci.end_line_no) {
         dest_line_no = p_RLine;
         return;
      }
   } while (false);

   // jump to the designated symbol
   _cb_goto_tag_in_file(cm.member_name, cm.file_name, cm.class_name, cm.type_name, cm.line_no, true);

}

